#!/usr/bin/env python3
import os,sys,glob,multiprocessing,time,csv,math,pprint,shutil
from parsl.app.app import python_app
from os.path import *
from mappertrac.subscripts import *

@python_app(executors=['worker'])
def run_freesurfer(params):

    input_dir = params['input_dir']
    sdir = params['work_dir']
    ID = params['ID']
    stdout = params['stdout']

    start_time = time.time()
    start_str = f'''
=====================================
{get_time_date()}
Started step 1: freesurfer
Arguments: 
{pprint.pformat(params, width=1)}
=====================================
'''
    write(stdout, start_str)
    print(start_str)

    input_dwi = join(input_dir, 'dwi', f'{ID}_dwi.nii.gz')
    input_bval = join(input_dir, 'dwi', f'{ID}_dwi.bval')
    input_bvec = join(input_dir, 'dwi', f'{ID}_dwi.bvec')
    input_T1 = join(input_dir, 'anat', f'{ID}_T1w.nii.gz')
    for _ in [input_dwi, input_bval, input_bvec, input_T1]:
        assert exists(_), f'Missing file {_}'
    
    smart_mkdir(sdir)
    work_dwi = join(sdir, 'hardi.nii.gz')
    work_bval = join(sdir, 'bvals')
    work_bvec = join(sdir, 'bvecs')
    work_T1 = join(sdir, 'T1.nii.gz')
    smart_copy(input_dwi, work_dwi)
    smart_copy(input_bval, work_bval)
    smart_copy(input_bvec, work_bvec)
    smart_copy(input_T1, work_T1)

    # Save NIfTI files in BIDS rawdata directory
    rawdata_dir = sdir.replace('/derivatives/', '/rawdata/')
    smart_copy(input_dir, rawdata_dir)

    ##################################
    # dti-preproc
    ##################################

    eddy_prefix = join(sdir, 'data_eddy')
    data_eddy = f'{eddy_prefix}.nii.gz'
    bet = join(sdir, 'data_bet.nii.gz')

    for _ in glob(f'{eddy_prefix}_tmp????.*') + glob(f'{eddy_prefix}_ref*'):
        smart_remove(_)

    run(f'fslroi {work_dwi} {eddy_prefix}_ref 0 1', params)
    run(f'fslsplit {work_dwi} {eddy_prefix}_tmp', params)

    timeslices = glob(f'{eddy_prefix}_tmp????.*')
    for _ in timeslices:
        run(f'flirt -in {_} -ref {eddy_prefix}_ref -nosearch -interp trilinear -o {_} -paddingsize 1', params)
    run(f'fslmerge -t {data_eddy} {" ".join(timeslices)}', params)
    run(f'bet {data_eddy} {bet} -m -f 0.3', params)

    ##################################
    # recon-all
    ##################################

    bet_mask = join(sdir, 'data_bet_mask.nii.gz')
    dti_params = join(sdir, 'DTIparams')
    dti_L1 = f'{dti_params}_L1.nii.gz'
    dti_L2 = f'{dti_params}_L2.nii.gz'
    dti_L3 = f'{dti_params}_L3.nii.gz'
    dti_MD = f'{dti_params}_MD.nii.gz'
    dti_RD = f'{dti_params}_RD.nii.gz'
    dti_MD = f'{dti_params}_MD.nii.gz'
    dti_AD = f'{dti_params}_AD.nii.gz'
    dti_FA = f'{dti_params}_FA.nii.gz'
    FA = join(sdir, 'FA.nii.gz')
    if exists(bet_mask):
        run(f'dtifit --verbose -k {data_eddy} -o {dti_params} -m {bet_mask} -r {work_bvec} -b {work_bval}', params)
        run(f'fslmaths {dti_L1} -add {dti_L2} -add {dti_L3} -div 3 {dti_MD}', params)
        run(f'fslmaths {dti_L2} -add {dti_L3} -div 2 {dti_RD}', params)
        smart_copy(dti_L1, dti_AD)
        smart_copy(dti_FA, FA)
    else:
        write(stdout, "Warning: failed to generate masked outputs")
        raise Exception(f"Failed BET step. Please check {stdout} for more info.")

    for _ in glob(f'{eddy_prefix}_tmp????.*') + glob(f'{eddy_prefix}_ref*'):
        smart_remove(_)

    # fsdir = join(sdir, 'freesurfer')
    # smart_mkdir(fsdir)
    # for _ in [work_dwi, work_bval, work_bvec, work_T1, bet_mask, dti_L1, dti_L2, dti_L3, dti_MD, dti_RD, dti_MD, dti_AD, dti_FA, FA]:
    #     smart_copy(_, join(fsdir, basename(_)))

    mri_out = join(sdir, 'mri', 'orig', '001.mgz')
    smart_mkdir(join(sdir, 'mri', 'orig'))
    run(f'mri_convert {work_T1} {mri_out}', params)

    ncores = int(multiprocessing.cpu_count())
    write(stdout, f'Running Freesurfer with {ncores} cores')
    run(f'recon-all -s . -all -notal-check -no-isrunning -parallel -openmp {ncores}', params)

    ##################################
    # mri_annotation2label
    ##################################
    mri_brain = join(sdir, 'mri', 'brain.mgz')
    mri_aseg = join(sdir, 'mri', 'aseg.mgz')
    aseg = join(sdir, 'aseg.nii.gz')
    bs = join(sdir, 'bs.nii.gz')
    FA2T1 = join(sdir, 'FA2T1.mat')
    T12FA = join(sdir, 'T12FA.mat')
    cort_label_dir = join(sdir, 'label_cortical')
    cort_vol_dir = join(sdir, 'volumes_cortical')
    cort_vol_dir_out = cort_vol_dir + '_s2fa'
    subcort_vol_dir = join(sdir, 'volumes_subcortical')
    subcort_vol_dir_out = subcort_vol_dir + '_s2fa'
    terminationmask = join(sdir, 'terminationmask.nii.gz')
    allvoxelscortsubcort = join(sdir, 'allvoxelscortsubcort.nii.gz')
    intersection = join(sdir, 'intersection.nii.gz')
    subcortical_index = [
        '10:lh_thalamus', '11:lh_caudate', '12:lh_putamen', '13:lh_pallidum', '17:lh_hippocampus', '18:lh_amygdala', '26:lh_acumbens', 
        '49:rh_thalamus', '50:rh_caudate', '51:rh_putamen', '52:rh_pallidum', '53:rh_hippocampus', '54:rh_amygdala', '58:rh_acumbens',
    ]
    EDI = join(sdir, 'EDI')
    EDI_allvols = join(EDI, 'allvols')

    smart_mkdir(cort_label_dir)
    smart_mkdir(cort_vol_dir)
    smart_mkdir(subcort_vol_dir)
    smart_mkdir(cort_vol_dir_out)
    smart_mkdir(subcort_vol_dir_out)
    smart_mkdir(EDI)
    smart_mkdir(EDI_allvols)

    run(f'mri_convert {mri_brain} {work_T1} ', params)
    run(f'mri_convert {mri_aseg} {aseg}', params)
    run(f'flirt -in {FA} -ref {work_T1} -omat {FA2T1}', params)
    run(f'convert_xfm -omat {T12FA} -inverse {FA2T1}', params)
    run(f'mri_annotation2label --subject . --hemi rh --annotation aparc --outdir {cort_label_dir}', params)
    run(f'mri_annotation2label --subject . --hemi lh --annotation aparc --outdir {cort_label_dir}', params)

    for label in glob(join(cort_label_dir, '*.label')):
        vol_file = join(cort_vol_dir, splitext(split(label)[1])[0] + '.nii.gz')
        run(f'mri_label2vol --label {label} --temp {work_T1} --identity --o {vol_file}', params)

    for line in subcortical_index:
        num = line.split(':')[0].lstrip().rstrip()
        area = line.split(':')[1].lstrip().rstrip()
        area_out = join(subcort_vol_dir, area + '.nii.gz')
        write(stdout, f'Processing {area}.nii.gz')
        run(f'fslmaths {aseg} -uthr {num} -thr {num} -bin {area_out}', params)

    for volume in glob(join(cort_vol_dir, '*.nii.gz')):
        out_vol = join(cort_vol_dir_out, splitext(splitext(split(volume)[1])[0])[0] + '_s2fa.nii.gz')
        write(stdout, f'Processing {split(volume)[1]} -> {split(out_vol)[1]}')
        run(f'flirt -in {volume} -ref {FA} -out {out_vol}  -applyxfm -init {T12FA}', params)
        run(f'fslmaths {out_vol} -thr 0.2 -bin {out_vol} ', params)

    for volume in glob(join(subcort_vol_dir, '*.nii.gz')):
        out_vol = join(subcort_vol_dir_out, splitext(splitext(split(volume)[1])[0])[0] + '_s2fa.nii.gz')
        write(stdout, f'Processing {split(volume)[1]} -> {split(out_vol)[1]}')
        run(f'flirt -in {volume} -ref {FA} -out {out_vol}  -applyxfm -init {T12FA}', params)
        run(f'fslmaths {out_vol} -thr 0.2 -bin {out_vol}', params)

    run(f'fslmaths {FA} -mul 0 {bs}', params)  # For now we fake a bs.nii.gz file
    maskseeds(sdir, join(cort_vol_dir + '_s2fa'), join(cort_vol_dir + '_s2fa_m'), 0.05, 1, 1, params)
    maskseeds(sdir, join(subcort_vol_dir + '_s2fa'), join(subcort_vol_dir + '_s2fa_m'), 0.05, 0.4, 0.4, params)
    saveallvoxels(sdir, join(cort_vol_dir + '_s2fa_m'), join(subcort_vol_dir + '_s2fa_m'), allvoxelscortsubcort, params)

    ##################################
    # termination mask
    ##################################
    smart_remove(terminationmask)
    run(f'fslmaths {FA} -uthr .15 {terminationmask}'.format(FA, terminationmask), params)
    run(f'fslmaths {terminationmask} -add {bs} {terminationmask}', params)
    run(f'fslmaths {terminationmask} -bin {terminationmask}', params)
    run(f'fslmaths {terminationmask} -mul {allvoxelscortsubcort} {intersection}', params)
    run(f'fslmaths {terminationmask} -sub {intersection} {terminationmask}', params)
    for file in glob(join(sdir, 'volumes_cortical_s2fa','*.nii.gz')):
        shutil.copy(file, EDI_allvols)
    for file in glob(join(sdir, 'volumes_subcortical_s2fa','*.nii.gz')):
        shutil.copy(file, EDI_allvols)
    validate(terminationmask, params)
    update_permissions(sdir, params)

    write(join(sdir, 'S1_COMPLETE'))
    
    finish_str = f'''
=====================================
{get_time_date()}
Finished step 1: freesurfer
Arguments: 
{pprint.pformat(params, width=1)}
Total time: {get_time_string(time.time() - start_time)} (HH:MM:SS)
=====================================
'''
    write(stdout, finish_str)
    print(finish_str)



#!/usr/bin/env python3
import os,sys,glob,multiprocessing,time,csv,math,pprint,shutil,platform,fcntl,errno,tempfile,json,psutil,random
import scipy.io
import numpy as np
from parsl.app.app import python_app
from os.path import *
from mappertrac.subscripts import *

def run_probtrackx(params):

    sdir = params['work_dir']
    assert exists(join(sdir, 'S1_COMPLETE')), 'Subject {sdir} must first run --freesurfer'
    assert exists(join(sdir, 'S2_COMPLETE')), 'Subject {sdir} must first run --bedpostx'

    pbtx_edges = get_edges_from_file(join(params['script_dir'], 'data/lists/list_edges_reduced.txt'))
    edges_per_chunk = 4
    n = edges_per_chunk
    edge_chunks = [pbtx_edges[i * n:(i + 1) * n] for i in range((len(pbtx_edges) + n - 1) // n )]

    start_future = start(params)
    process_futures = []
    for edge_chunk in edge_chunks:
        process_futures.append(process(params, edge_chunk, inputs=[start_future]))
    return combine(params, inputs=process_futures)

@python_app(executors=['worker'])
def start(params, inputs=[]):

    sdir = params['work_dir']
    stdout = params['stdout']

    start_time = time.time()
    start_str = f'''
=====================================
{get_time_date()}
Started step 3: probtrackx
Arguments: 
{pprint.pformat(params, width=1)}
=====================================
'''
    write(stdout, start_str)
    print(start_str)
    time_log = join(sdir, 'start_time_s3.txt')
    smart_remove(time_log)
    write(time_log, start_time)

    sdir = params['work_dir']
    output_dir = params['output_dir']
    pbtk_dir = join(sdir,"EDI","PBTKresults")
    connectome_dir = join(sdir,"EDI","CNTMresults")
    derivatives_dir_tmp = join(output_dir, 'derivatives', "tmp")
    sdir_tmp = join(sdir, "tmp")
    smart_remove(pbtk_dir)
    smart_remove(connectome_dir)
    smart_remove(sdir_tmp)
    smart_mkdir(pbtk_dir)
    smart_mkdir(connectome_dir)
    smart_mkdir(sdir_tmp)
    time.sleep(random.randrange(0, 10)) # random sleep to avoid parallel collision
    smart_mkdir(derivatives_dir_tmp)

@python_app(executors=['worker'])
def process(params, edges, inputs=[]):

    sdir = params['work_dir']
    stdout = params['stdout']
    output_dir = params['output_dir']
    pbtx_sample_count = params['pbtx_sample_count']
    derivatives_dir_tmp = join(output_dir, 'derivatives', "tmp")
    sdir_tmp = join(sdir, "tmp")
    EDI_allvols = join(sdir,"EDI","allvols")
    pbtk_dir = join(sdir,"EDI","PBTKresults")
    connectome_dir = join(sdir,"EDI","CNTMresults")
    bedpostxResults = join(sdir,"bedpostx_b1000.bedpostX")
    merged = join(bedpostxResults,"merged")
    nodif_brain_mask = join(bedpostxResults,"nodif_brain_mask.nii.gz")
    allvoxelscortsubcort = join(sdir,"allvoxelscortsubcort.nii.gz")
    terminationmask = join(sdir,"terminationmask.nii.gz")
    bs = join(sdir,"bs.nii.gz")

    ##################################
    # Memory Management
    ##################################

    pbtx_max_memory = psutil.virtual_memory().total * 1.0E-9
    node_name = platform.uname().node.strip()
    assert node_name and ' ' not in node_name, "Invalid node name {}".format(node_name)
    mem_record = join(derivatives_dir_tmp, node_name + '.json') # Keep record to avoid overusing node memory

    # Only access mem_record with file locking to avoid outdated data
    def open_mem_record(mode = 'r'):
        f = None
        while True:
            try:
                f = open(mem_record, mode, newline='')
                fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                break
            except IOError as e:
                # raise on unrelated IOErrors
                if e.errno != errno.EAGAIN:
                    raise
                else:
                    time.sleep(0.1)
        assert f is not None, "Failed to open mem_record {}".format(mem_record)
        return f

    def estimate_total_memory_usage():
        f = open_mem_record('r')
        mem_dict = json.load(f)
        fcntl.flock(f, fcntl.LOCK_UN)
        f.close()
        mem_sum = 0.0
        for task_mem in mem_dict.values():
            mem_sum += float(task_mem)
        return mem_sum

    def estimate_task_mem_usage():
        total_size = 0
        total_size += os.path.getsize(allvoxelscortsubcort)
        total_size += os.path.getsize(terminationmask)
        total_size += os.path.getsize(bs)
        for dirpath, dirnames, filenames in os.walk(bedpostxResults):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if not os.path.islink(fp):
                    total_size += os.path.getsize(fp)

        max_region_size = 0
        for edge in edges:
            a, b = edge
            a_file = join(EDI_allvols, a + "_s2fa.nii.gz")
            b_file = join(EDI_allvols, b + "_s2fa.nii.gz")
            a_size = os.path.getsize(a_file)
            b_size = os.path.getsize(b_file)
            max_region_size = max([a_size, b_size, max_region_size])
        total_size += max_region_size
        return float(total_size) * 1.0E-9

    def add_task():
        task_id = '0'
        f = open_mem_record('r')
        if not exists(mem_record):
            json.dump({task_id:task_mem_usage}, f)
        else:
            mem_dict = json.load(f)
            task_ids = [int(x) for x in mem_dict.keys()] + [0] # append zero in case task_ids empty
            task_id = str(max(task_ids) + 1) # generate incremental task_id
            mem_dict[task_id] = task_mem_usage
            tmp_fp, tmp_path = tempfile.mkstemp(dir=sdir_tmp)
            with open(tmp_path, 'w', newline='') as tmp: # file pointer not consistent, so we open using the pathname
                json.dump(mem_dict, tmp)
            os.replace(tmp_path, mem_record) # atomic on POSIX systems. flock is advisory, so we can still overwrite.
        fcntl.flock(f, fcntl.LOCK_UN)
        f.close()
        return task_id

    def remove_task(task_id):
        f = open_mem_record('r')
        mem_dict = json.load(f)
        mem_dict.pop(task_id, None)
        tmp_fp, tmp_path = tempfile.mkstemp(dir=sdir_tmp)
        with open(tmp_path, 'w', newline='') as tmp:
            json.dump(mem_dict, tmp)
        os.replace(tmp_path, mem_record)
        fcntl.flock(f, fcntl.LOCK_UN)
        f.close()

    sleep_timeout = 7200

    task_mem_usage = estimate_task_mem_usage()
    assert task_mem_usage < pbtx_max_memory, f'Task consumes more memory ({task_mem_usage:.2f} GB) than available ({pbtx_max_memory:.2f} GB)'
    total_sleep = 0
    # Memory record is atomic, but might not be updated on time
    # So we randomize sleep to discourage multiple tasks hitting at once
    init_sleep = random.randrange(0, 30)
    write(stdout, "Sleeping for {:d} seconds".format(init_sleep))
    total_sleep += init_sleep
    time.sleep(init_sleep)

    if not exists(mem_record):
        f = open_mem_record('w')
        json.dump({}, f)
        fcntl.flock(f, fcntl.LOCK_UN)
        f.close()

    total_mem_usage = estimate_total_memory_usage()
    # Then we sleep until memory usage is low enough
    while total_mem_usage + task_mem_usage > pbtx_max_memory:
        sleep_interval = random.randrange(5, 60)
        write(stdout, "Sleeping for {:d} seconds. Memory usage: {:.2f}/{:.2f} GB".format(sleep_interval, total_mem_usage, pbtx_max_memory))
        total_sleep += sleep_interval
        if total_sleep > sleep_timeout:
            raise Exception('Retrying task that has slept longer than 2 hours')
        time.sleep(sleep_interval)
        total_mem_usage = estimate_total_memory_usage()
    write(stdout, "Running Probtrackx after sleeping for {} seconds".format(total_sleep))

    # Insert task and memory usage into record
    task_id = add_task()

    ##################################
    # Tractography
    ##################################

    try:
        for edge in edges:
            a, b = edge
            a_file = join(EDI_allvols, a + "_s2fa.nii.gz")
            b_file = join(EDI_allvols, b + "_s2fa.nii.gz")
            tmp = join(sdir, "tmp", "{}_to_{}".format(a, b))
            a_to_b_formatted = "{}_s2fato{}_s2fa.nii.gz".format(a,b)
            a_to_b_file = join(pbtk_dir,a_to_b_formatted)
            waypoints = join(tmp,"waypoint.txt")
            waytotal = join(tmp, "waytotal")
            assert exists(a_file) and exists(b_file), "Error: Both Freesurfer regions must exist: {} and {}".format(a_file, b_file)
            smart_remove(a_to_b_file)
            smart_remove(tmp)
            smart_mkdir(tmp)
            write(stdout, "Running subproc: {} to {}".format(a, b))
            write(waypoints, b_file.replace(sdir, "/mnt"))

            exclusion = join(tmp,"exclusion.nii.gz")
            termination = join(tmp,"termination.nii.gz")
            run("fslmaths {} -sub {} {}".format(allvoxelscortsubcort, a_file, exclusion), params)
            run("fslmaths {} -sub {} {}".format(exclusion, b_file, exclusion), params)
            run("fslmaths {} -add {} {}".format(exclusion, bs, exclusion), params)
            run("fslmaths {} -add {} {}".format(terminationmask, b_file, termination), params)

            pbtx_args = (" -x {} ".format(a_file) +
                # " --pd -l -c 0.2 -S 2000 --steplength=0.5 -P 1000" +
                " --pd -l -c 0.2 -S 2000 --steplength=0.5 -P {}".format(pbtx_sample_count) +
                " --waypoints={} --avoid={} --stop={}".format(waypoints, exclusion, termination) +
                " --forcedir --opd --rseed={}".format(random.randint(1000,9999)) +
                " -s {}".format(merged) +
                " -m {}".format(nodif_brain_mask) +
                " --dir={}".format(tmp) +
                " --out={}".format(a_to_b_formatted)
                )
            run("probtrackx2" + pbtx_args, params)

            waytotal_count = 0
            if exists(waytotal):
                with open(waytotal, 'r') as f:
                    waytotal_count = f.read().strip()
                    fdt_count = run("fslmeants -i {} -m {} | head -n 1".format(join(tmp, a_to_b_formatted), b_file), params) # based on getconnectome script
                    if not is_float(waytotal_count):
                        write(stdout, "Error: Failed to read waytotal_count value {} in {}".format(waytotal_count, edge))
                        continue
                    if not is_float(fdt_count):
                        write(stdout, "Error: Failed to read fdt_count value {} in {}".format(fdt_count, edge))
                        continue
                    edge_file = join(connectome_dir, "{}_to_{}.dot".format(a, b))
                    smart_remove(edge_file)
                    write(edge_file, "{} {} {} {}".format(a, b, waytotal_count, fdt_count))

                    # Error check edge file
                    with open(edge_file) as f:
                        line = f.read().strip()
                        if len(line) > 0: # ignore empty lines
                            chunks = [x.strip() for x in line.split(' ') if x]
                            if not (len(chunks) == 4 and is_float(chunks[2]) and is_float(chunks[3])):
                                write(stdout, "Error: Connectome {} has invalid edge {} to {}".format(edge_file, a, b))
                                continue
            else:
                write(stdout, 'Error: failed to find waytotal for {} to {}'.format(a, b))
            copyfile(join(tmp, a_to_b_formatted), a_to_b_file) # keep edi output
            if not a == "lh.paracentral": # discard all temp files except these for debugging
                smart_remove(tmp)
    finally:
        remove_task(task_id)

@python_app(executors=['worker'])
def combine(params, inputs=[]):

    sdir = params['work_dir']
    stdout = params['stdout']
    pbtx_sample_count = params['pbtx_sample_count']
    pbtx_edges = get_edges_from_file(join(params['script_dir'], 'data/lists/list_edges_tiny.txt'))
    connectome_idx_list = join(params['script_dir'], 'data/lists/connectome_idxs.txt')
    start_time = time.time()
    connectome_dir = join(sdir,"EDI","CNTMresults")
    oneway_list = join(sdir, "connectome_{}samples_oneway.txt".format(pbtx_sample_count))
    twoway_list = join(sdir, "connectome_{}samples_twoway.txt".format(pbtx_sample_count))
    oneway_nof = join(sdir, "connectome_{}samples_oneway_nof.mat".format(pbtx_sample_count)) # nof = number of fibers
    twoway_nof = join(sdir, "connectome_{}samples_twoway_nof.mat".format(pbtx_sample_count))
    oneway_nof_normalized = join(sdir, "connectome_{}samples_oneway_nofn.mat".format(pbtx_sample_count)) # nofn = number of fibers, normalized
    twoway_nof_normalized = join(sdir, "connectome_{}samples_twoway_nofn.mat".format(pbtx_sample_count))
    pbtk_dir = join(sdir,"EDI","PBTKresults")
    consensus_dir = join(pbtk_dir,"twoway_consensus_edges")
    edi_maps = join(sdir,"EDI","EDImaps")
    edge_total = join(edi_maps,"FAtractsumsTwoway.nii.gz")
    tract_total = join(edi_maps,"FAtractsumsRaw.nii.gz")
    smart_remove(oneway_list)
    smart_remove(twoway_list)
    smart_remove(oneway_nof_normalized)
    smart_remove(twoway_nof_normalized)
    smart_remove(oneway_nof)
    smart_remove(twoway_nof)
    smart_remove(edi_maps)
    smart_mkdir(pbtk_dir)
    smart_mkdir(consensus_dir)
    smart_mkdir(edi_maps)
    oneway_edges = {}
    twoway_edges = {}

    consensus_edges = []
    for edge in pbtx_edges:
        a, b = edge
        if [a, b] in consensus_edges or [b, a] in consensus_edges:
            continue
        consensus_edges.append(edge)

    copyfile(connectome_idx_list, join(sdir, 'connectome_idxs.txt')) # give each subject a copy for reference

    ##################################
    # Compile connectome matrices
    ##################################
    vol_idxs = {}
    with open(connectome_idx_list) as f:
        lines = [x.strip() for x in f.readlines() if x]
        max_idx = -1
        for line in lines:
            vol, idx = line.split(',', 1)
            idx = int(idx)
            vol_idxs[vol] = idx
            if idx > max_idx:
                max_idx = idx
        oneway_nof_normalized_matrix = np.zeros((max_idx+1, max_idx+1))
        oneway_nof_matrix = np.zeros((max_idx+1, max_idx+1))
        twoway_nof_normalized_matrix = np.zeros((max_idx+1, max_idx+1))
        twoway_nof_matrix = np.zeros((max_idx+1, max_idx+1))

    for edge in pbtx_edges:
        a, b = edge
        edge_file = join(connectome_dir, "{}_to_{}.dot".format(a, b))
        with open(edge_file) as f:
            chunks = [x.strip() for x in f.read().strip().split(' ') if x]
            a_to_b = (chunks[0], chunks[1])
            b_to_a = (chunks[1], chunks[0])
            waytotal_count = float(chunks[2])
            fdt_count = float(chunks[3])
            if b_to_a in twoway_edges:
                twoway_edges[b_to_a][0] += waytotal_count
                twoway_edges[b_to_a][1] += fdt_count
            else:
                twoway_edges[a_to_b] = [waytotal_count, fdt_count]
            oneway_edges[a_to_b] = [waytotal_count, fdt_count]

    for a_to_b in oneway_edges:
        a = a_to_b[0]
        b = a_to_b[1]
        for vol in a_to_b:
            if vol not in vol_idxs:
                write(stdout, 'Error: could not find {} in connectome idxs'.format(vol))
                break
        else:
            write(oneway_list, "{} {} {} {}".format(a, b, oneway_edges[a_to_b][0], oneway_edges[a_to_b][1]))
            oneway_nof_matrix[vol_idxs[a]][vol_idxs[b]] = oneway_edges[a_to_b][0]
            oneway_nof_normalized_matrix[vol_idxs[a]][vol_idxs[b]] = oneway_edges[a_to_b][1]

    for a_to_b in twoway_edges:
        a = a_to_b[0]
        b = a_to_b[1]
        for vol in a_to_b:
            if vol not in vol_idxs:
                write(stdout, 'Error: could not find {} in connectome idxs'.format(vol))
                break
        else:
            write(twoway_list, "{} {} {} {}".format(a, b, twoway_edges[a_to_b][0], twoway_edges[a_to_b][1]))
            twoway_nof_matrix[vol_idxs[a]][vol_idxs[b]] = twoway_edges[a_to_b][0]
            twoway_nof_normalized_matrix[vol_idxs[a]][vol_idxs[b]] = twoway_edges[a_to_b][1]
    scipy.io.savemat(oneway_nof, {'data': oneway_nof_matrix})
    scipy.io.savemat(oneway_nof_normalized, {'data': oneway_nof_normalized_matrix})
    scipy.io.savemat(twoway_nof, {'data': twoway_nof_matrix})
    scipy.io.savemat(twoway_nof_normalized, {'data': twoway_nof_normalized_matrix})
    smart_copy(twoway_nof_normalized, join(dirname(sdir), basename(twoway_nof_normalized)))
    smart_copy(twoway_list, join(dirname(sdir), basename(twoway_list)))
    
    ##################################
    # EDI consensus
    ##################################
    for edge in pbtx_edges:
        a, b = edge
        a_to_b = "{}_to_{}".format(a, b)
        a_to_b_file = join(pbtk_dir,"{}_s2fato{}_s2fa.nii.gz".format(a,b))
        b_to_a_file = join(pbtk_dir,"{}_s2fato{}_s2fa.nii.gz".format(b,a))
        if not exists(a_to_b_file):
            write(stdout, "Error: cannot find {}".format(a_to_b_file))
            return
        if not exists(b_to_a_file):
            write(stdout, "Error: cannot find {}".format(b_to_a_file))
            return
        consensus = join(consensus_dir, a_to_b + '.nii.gz')
        amax = run('fslstats {} -R | cut -f 2 -d \\" \\" '.format(a_to_b_file), params).strip()
        if not is_float(amax):
            write(stdout, "Error: fslstats on {} returns invalid value {}".format(a_to_b_file, amax))
            return
        amax = int(float(amax))
        bmax = run('fslstats {} -R | cut -f 2 -d \\" \\" '.format(b_to_a_file), params).strip()
        if not is_float(bmax):
            write(stdout, "Error: fslstats on {} returns invalid value {}".format(b_to_a_file, bmax))
            return
        bmax = int(float(bmax))
        write(stdout, "amax = {}, bmax = {}".format(amax, bmax))
        if amax > 0 and bmax > 0:
            tmp1 = join(pbtk_dir, "{}_to_{}_tmp1.nii.gz".format(a, b))
            tmp2 = join(pbtk_dir, "{}_to_{}_tmp2.nii.gz".format(b, a))
            run("fslmaths {} -thrP 5 -bin {}".format(a_to_b_file, tmp1), params)
            run("fslmaths {} -thrP 5 -bin {}".format(b_to_a_file, tmp2), params)
            run("fslmaths {} -add {} -thr 1 -bin {}".format(tmp1, tmp2, consensus), params)
            smart_remove(tmp1)
            smart_remove(tmp2)
        else:
            with open(join(pbtk_dir, "zerosl.txt"), 'a') as log:
                log.write("For edge {}:\n".format(a_to_b))
                log.write("{} is thresholded to {}\n".format(a, amax))
                log.write("{} is thresholded to {}\n".format(b, bmax))

    # Collect number of probtrackx tracts per voxel
    for edge in pbtx_edges:
        a, b = edge
        a_to_b_formatted = "{}_s2fato{}_s2fa.nii.gz".format(a,b)
        a_to_b_file = join(pbtk_dir,a_to_b_formatted)
        if not exists(tract_total):
            copyfile(a_to_b_file, tract_total)
        else:
            run("fslmaths {0} -add {1} {1}".format(a_to_b_file, tract_total), params)

    # Collect number of parcel-to-parcel edges per voxel
    for edge in consensus_edges:
        a, b = edge
        consensus = join(consensus_dir, "{}_to_{}.nii.gz".format(a,b))
        if not exists(consensus):
            write(stdout,"{} has been thresholded. See {} for details".format(edge, join(pbtk_dir, "zerosl.txt")))
            continue
        if not exists(edge_total):
            copyfile(consensus, edge_total)
        else:
            run("fslmaths {0} -add {1} {1}".format(consensus, edge_total), params)
    if not exists(edge_total):
        write(stdout, "Error: Failed to generate {}".format(edge_total))
    else:
        smart_copy(edge_total, join(dirname(sdir), 'EDI_' + basename(edge_total)))

    update_permissions(sdir, params)
    write(join(sdir, 'S3_COMPLETE'))
    
    time_log = join(sdir, 'start_time_s3.txt')
    with open(time_log) as f:
        start_time = float(f.read())
    finish_str = f'''
=====================================
{get_time_date()}
Finished step 3: probtrackx
Arguments: 
{pprint.pformat(params, width=1)}
Total time: {get_time_string(time.time() - start_time)} (HH:MM:SS)
=====================================
'''
    write(stdout, finish_str)
    print(finish_str)



import time,datetime,os,subprocess,sys,shutil,hashlib,grp,mmap,fnmatch,gzip,re,random
import distutils.dir_util
from glob import glob
from os.path import exists,join,split,splitext,abspath,basename,dirname,isdir,samefile,getsize
from shutil import copyfile,copytree,rmtree,ignore_patterns
from os import system,environ,makedirs,remove
from subprocess import Popen,PIPE
from itertools import islice

def smart_mkdir(path):
    if exists(path):
        return
    if not isdir(path):
        makedirs(path)

def smart_remove(path):
    """Remove all files and directories if they exist
    """
    if isdir(path):
        rmtree(path)
    elif exists(path):
        try:
            remove(path)
        except OSError:
            pass

def smart_copy(src, dest, exclude=[]):
    """Copy file or directory, while ignoring non-existent or equivalent files
    """
    if exists(dest) and samefile(src, dest):
        print("Warning: ignoring smart_copy because src and dest both point to {}".format(dest))
        return
    if not exists(dirname(dest)):
        smart_mkdir(dirname(dest))
    if isdir(src):
        tmp_root = join(dirname(dirname(__file__)), 'tmp')
        tmp = join(tmp_root, f'{basename(src)}_{random.randint(0,1000)}')
        smart_remove(tmp)
        copytree(src, tmp, ignore=ignore_patterns(*exclude))   # copy tree with excludes
        distutils.dir_util.copy_tree(tmp, dest)                # then copy tree without overwrite
        smart_remove(tmp)
    else:
        for pattern in exclude:
            if fnmatch.fnmatch(src, pattern):
                print('Did not copy {} because of exclude={}'.format(src, exclude))
                return
        copyfile(src, dest)

def exist_all(paths):
    for path in paths:
        if not exists(path):
            return False
    return True

def run(command, params=None, ignore_errors=False, print_output=True, print_time=False, working_dir=None):
    """Run a command in a subprocess.
    Safer than raw execution. Can also write to logs and utilize a container.
    """
    start = int(time.time())
    work_dir = params['work_dir']
    stdout = params['stdout'] if (params and 'stdout' in params) else None
    container = params['container'] if (params and 'container' in params) else None
    use_gpu = params['use_gpu'] if (params and 'use_gpu' in params) else None
    container_cwd = params['container_cwd'] if (params and 'container_cwd' in params) else None

    # When using a container, change all paths to be relative to its mounted directory (hideous, but works without changing other code)
    if container is not None:
        command = command.replace(work_dir, "/mnt")
        command = (f'singularity exec {"--nv" if use_gpu else ""} ' +
            f'--cleanenv ' +
            f'--home /fake_home_dir ' +
            f'-B {work_dir}:/mnt {container} ' +
            f'sh -c "{command}"')
        print(command)
        if container_cwd:
            command = "cd {}; {}".format(container_cwd, command)
        if stdout:
            write(stdout, command)

    process = Popen(command, stdout=PIPE, stderr=subprocess.STDOUT, shell=True, env=environ, cwd=working_dir)
    line = ""
    while True:
        new_line = process.stdout.readline()
        new_line = str(new_line, 'utf-8')[:-1]
        if print_output and new_line:
            print(new_line)
        if stdout and not new_line.isspace():
            write(stdout, new_line)
        if new_line == '' and process.poll() is not None:
            break
        line = new_line
    if process.returncode != 0 and not ignore_errors:
        if stdout and not new_line.isspace():
            write(stdout, "Error: non zero return code")
            write(stdout, get_time_date())
        raise Exception("Non zero return code: {}\nCommand: {}".format(process.returncode, command))
    else:
        tokens = command.split(' ')
        if print_time:
            print("Running {} took {} (h:m:s)".format(tokens[0], get_time_string(int(time.time()) - start)))
            if len(tokens) > 1:
                print("\tArgs: {}".format(' '.join(tokens[1:])))
    return line  # return the last output line

def is_float(value):
    try:
        float(value)
        return True
    except ValueError:
        return False

def is_integer(value):
    try:
        int(value)
        return True
    except ValueError:
        return False

def clamp(value, lo, hi):
    return max(lo, min(value, hi))

def str2bool(string):
    if string is None:
        return string
    return str(string).lower() in ("yes", "true", "t", "1")

def get_time_date():
    """Returns date and time as Y-m-d H:M am/pm
    """
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M %p")

def get_time_string(seconds):
    """Returns seconds as Slurm-compatible time string
    """
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    time_string = "{:02d}:{:02d}:{:02d}".format(int(h), int(m), int(s))
    if h > 99999 or h < 0:
        return "00:00:00"
    return time_string

def get_time_seconds(string):
    """Returns Slurm-compatible time string as seconds
    """
    while (len(string.split(":")) < 3):
        string = "00:" + string
    return sum(secs * int(digit) for secs,digit in zip([3600, 60, 1], string.split(":")))

def get_start(function_name):
    return "Starting {} at {}\n".format(basename(str(function_name)), get_time_date())

def get_finish(function_name):
    return "Finished {} at {}\n".format(basename(str(function_name)), get_time_date())

def print_start():
    print(get_start())
    return time.time()

def print_finish(start_time):
    print(get_finish())

def write(path, output='', params={}):
    if params and 'container' in params and 'work_dir' in params:
        command = command.replace(params['work_dir'], "/mnt")
    # make path to file if not an empty string
    if dirname(path):
        smart_mkdir(dirname(path))
    with open(path, 'a') as f:
        f.write(str(output) + "\n")

def write_error(path, output, params={}):
    write(path, 'Exception: ' + output, params)
    raise output

def update_permissions(directory, params):
    """Give user and group permissions to all generated files.
    """
    start_time = time.time()
    stdout = params['stdout']
    run("find {} -type f -print0 | xargs -0 -I _ chmod 770 _".format(directory), params)
    run("find {} -type d -print0 | xargs -0 -I _ chmod 2770 _".format(directory), params)
    if 'group' in params:
        group = params['group']
        run("find {} -type f -print0 | xargs -0 -I _ chgrp {} _".format(directory, group), params)
        run("find {} -type d -print0 | xargs -0 -I _ chgrp {} _".format(directory, group), params)
    write(stdout, "Updated file permissions, took {} (h:m:s)".format(get_time_string(time.time() - start_time)))

def generate_checksum(input_dir):
    """Return checksum of subject input files. This ensures re-computation if inputs change.
    """
    buf_size = 65536  # read file in 64kb chunks
    md5 = hashlib.md5()
    for fname in ['anat.nii.gz','bvals','bvecs','hardi.nii.gz']:
        f = open(join(input_dir, fname), 'rb')
        while True:
            data = f.read(buf_size)
            if not data:
                break
            md5.update(data)
        f.close()
    return md5.hexdigest()

def get_log_path(template):
    """Return paths for log outputs.
    """
    path, ext = splitext(template)
    idx = 0
    new_log = path + "_{:02d}".format(idx) + ext
    while exists(new_log):
        idx += 1
        if idx >= 100:
            raise Exception("Could not find valid filepath for template {}".format(template))
        new_log = path + "_{:02d}".format(idx) + ext
    prev_log = path + "_{:02d}".format(idx-1) + ext
    return new_log, prev_log, idx

def is_log_complete(path):
    if exists(path):
        with open(path, 'rb', 0) as f, mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as s:
            return s.find(b'stdout_log_complete') != -1

def running_step(steps, *argv):
    """Return true if running any step in arguments.
    """
    for step in argv:
        if step in steps:
            return True
    return False

def copy_dir(src, dest):
    if not isdir(src):
        raise Exception("Source directory {} does not exist".format(src))
    smart_mkdir(dest)
    run("cp -Rf {} {}".format(join(src,"."), join(dest,".")))

def parse_default(arg, default, args_obj, pending_args):
    if not hasattr(args_obj, arg) or getattr(args_obj, arg) is None:
        setattr(args_obj, arg, default)
    if isinstance(default, bool):
        setattr(args_obj, arg, str2bool(getattr(args_obj, arg)))
    elif isinstance(default, str):
        if str(getattr(args_obj, arg)).lower() == "none":
            setattr(args_obj, arg, None)
    pending_args.pop(arg, None)

def add_binary_vol(src, target, params={}):
    run("fslmaths {} -add {} {}".format(src, target, target), params)
    run("fslmaths {} -bin {}".format(target, target), params)

def sub_binary_vol(src, target, params={}):
    intersection = "intersection.nii.gz"
    run("fslmaths {} -mul {} {}".format(src, target, intersection), params)
    run("fslmaths {} -sub {} {}".format(target, intersection, target), params)
    run("fslmaths {} -bin {}".format(target, target), params)

def strip_trailing_slash(path):
    if path.endswith('/') or path.endswith('\\'):
        path = path[:-1]
    return path

def get_edges_from_file(file):
    edges = []
    with open(file) as f:
        for edge in f.readlines():
            if edge.isspace():
                continue
            edges.append(edge.replace("_s2fa", "").strip().split(',', 1))
    return edges

def generate_edge_list(vol_dir, path='lists/listEdgesEDIAll.txt'):
    """Not used during runtime. Generates a list of all possible edges from Freesurfer output.
    """
    with open(path,'w') as f:
        files = glob(join(vol_dir,"*_s2fa.nii.gz"))  # Assemble all files
        files = [abspath(vol) for vol in files]
        for a in files:
            for b in files:
                if a != b:
                    a1 = splitext(splitext(split(a)[1])[0])[0]
                    b1 = splitext(splitext(split(b)[1])[0])[0]
                    f.write("{},{}\n".format(basename(a1),basename(b1)))

def compress_file(file):
    compressed_file = file + '.gz'
    with open(file, 'rb') as f_in:
        with gzip.open(compressed_file, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    return compressed_file

def get_bids_subject_name(sname):
    sname = sname.replace('sub-', '') # make naming consistent
    regex = re.compile('[^a-zA-Z0-9]')
    return 'sub-{}'.format(regex.sub('', sname))

def validate(file, params={}):
    file = file.strip()
    if not file.endswith('.nii.gz'):
        file = file + '.nii.gz'
    mean = run("fslstats {} -m | head -n 1".format(file), params)
    assert is_float(mean), "Invalid mean value in {}".format(file)
    assert float(mean) != 0, "Zero mean value in {}".format(file)

def maskseeds(root_dir,input_dir,output_dir,low_threshold,high_threshold,high_threshold_thalamus,params):
    
    smart_remove(output_dir)
    smart_mkdir(output_dir)

    tmp_thalamus = join(root_dir, "tmp_thalamus.nii.gz")
    tmp = join(root_dir, "tmp.nii.gz")
        
    # Now create two transformed volumes with threshold 1 and 2
    run("fslmaths {} -thr {} -uthr {} -bin {}".format(join(root_dir,"FA.nii.gz"),low_threshold,high_threshold,tmp), params)
    run("fslmaths {} -thr {} -uthr {} -bin {}".format(join(root_dir,"FA.nii.gz"),low_threshold,high_threshold_thalamus,tmp_thalamus), params)
    
    for seed in glob(join(input_dir,"*s2fa.nii.gz")):

        region = split(seed)[1].split(".")[1].split("_")[0]
    
        if region == "thalamus":
            run("fslmaths {} -mas {} {}".format(seed, tmp_thalamus, join(output_dir,split(seed)[1])), params)
        else:
            run("fslmaths {} -mas {} {}".format(seed, tmp, join(output_dir,split(seed)[1])), params)
           
    smart_remove(tmp_thalamus)
    smart_remove(tmp)
    
def saveallvoxels(root_dir,cortical_dir,subcortical_dir,output_name,params):
    
    smart_remove(join(root_dir,"cort.nii.gz"))
    smart_remove(join(root_dir,"subcort.nii.gz"))
    
    all_vols = ""
    for vol in glob(join(cortical_dir,"*_s2fa.nii.gz")):
        all_vols += " " + vol
    
    run("find_the_biggest {} {}".format(all_vols,join(root_dir,"cort.nii.gz")), params)
        
    all_vols = ""
    for vol in glob(join(subcortical_dir,"*_s2fa.nii.gz")):
        all_vols += " " + vol
    
    run("find_the_biggest {} {}".format(all_vols,join(root_dir,"subcort.nii.gz")), params)
    run("fslmaths {} -add {} {} ".format(join(root_dir,"cort.nii.gz"),join(root_dir,"subcort.nii.gz"),output_name), params)
    run("fslmaths {} -bin {}".format(output_name,output_name), params)

def append_to_filename(filename, append_string):
    name, ext = os.path.splitext(filename)
    return "{}_{}{}".format(name, append_string, ext)

def deinterlace(sequence_file, forward_file, reverse_file):

    if sequence_file.endswith('.gz'):
        run(f"gzip -k -d {sequence_file}")
        sequence_file = sequence_file[:-3]

    smart_remove(forward_file)
    smart_remove(reverse_file)
    smart_remove(forward_file + '.gz')
    smart_remove(reverse_file + '.gz')

    i = 0
    forward = True
    file_size = getsize(sequence_file)
    chunk_size = 0
    with open(sequence_file) as f1:
        while True:
            lines = list(islice(f1, 4))
            if not lines:
                break
            with open(forward_file if forward else reverse_file, 'a') as f2:
                f2.write(''.join(lines))
            i += 1
            forward = not forward

            if chunk_size == 0:
                chunk_size = len(''.join(lines).encode('ascii')) - 1
            if i % 1000 == 0:
                print("Deinterlacing {}, {:.2%}...".format(sequence_file, chunk_size * i / file_size))
    
    run(f"gzip {forward_file}")
    run(f"gzip {reverse_file}")

    smart_remove(sequence_file)

def replace_extension(path, extension=''):
    dirname, basename = split(path)
    return join(dirname, basename.split('.')[0] + extension)



#!/usr/bin/env python3
import os,sys,glob,multiprocessing,time,csv,math,pprint,shutil
import GPUtil
from parsl.app.app import python_app
from os.path import *
from mappertrac.subscripts import *

@python_app(executors=['worker'])
def run_bedpostx(params):

    input_dir = params['input_dir']
    sdir = params['work_dir']
    ID = params['ID']
    stdout = params['stdout']

    assert exists(join(sdir, 'S1_COMPLETE')), 'Subject {sdir} must first run --freesurfer'
    assert len(GPUtil.getAvailable(includeNan=True)) > 0, 'Failed to find CUDA-capable GPU'

    start_time = time.time()
    start_str = f'''
=====================================
{get_time_date()}
Started step 2: bedpostx
Arguments: 
{pprint.pformat(params, width=1)}
=====================================
'''
    write(stdout, start_str)
    print(start_str)

    bedpostx = join(sdir,'bedpostx_b1000')
    bedpostxResults = join(sdir,'bedpostx_b1000.bedpostX')
    th1 = join(bedpostxResults, 'merged_th1samples')
    ph1 = join(bedpostxResults, 'merged_ph1samples')
    th2 = join(bedpostxResults, 'merged_th2samples')
    ph2 = join(bedpostxResults, 'merged_ph2samples')
    dyads1 = join(bedpostxResults, 'dyads1')
    dyads2 = join(bedpostxResults, 'dyads2')
    brain_mask = join(bedpostxResults, 'nodif_brain_mask')
    smart_remove(bedpostxResults)
    smart_mkdir(bedpostx)
    smart_mkdir(bedpostxResults)
    smart_copy(join(sdir, 'data_eddy.nii.gz'), join(bedpostx, 'data.nii.gz'))
    smart_copy(join(sdir, 'data_bet_mask.nii.gz'), join(bedpostx, 'nodif_brain_mask.nii.gz'))
    smart_copy(join(sdir, 'bvals'), join(bedpostx, 'bvals'))
    smart_copy(join(sdir, 'bvecs'), join(bedpostx, 'bvecs'))

    bedpostx_sh = join(sdir, 'bedpostx.sh')
    smart_remove(bedpostx_sh)
    write(bedpostx_sh, 'export CUDA_LIB_DIR=$CUDA_8_LIB_DIR\n' +
                       'export LD_LIBRARY_PATH=$CUDA_LIB_DIR:$LD_LIBRARY_PATH\n' +
                       'bedpostx_gpu /mnt/bedpostx_b1000 -NJOBS 4')
    gpu_params = params.copy()
    gpu_params['use_gpu'] = True
    run('sh ' + bedpostx_sh, gpu_params)
    
    # hacky validation step
    with open(stdout) as f:
        log_content = f.read()
        for i in range(1, 5):
            assert('{:d} parts processed out of 4'.format(i) in log_content)

    run(f'make_dyadic_vectors {th1} {ph1} {brain_mask} {dyads1}', params)
    run(f'make_dyadic_vectors {th2} {ph2} {brain_mask} {dyads2}', params)
    update_permissions(sdir, params)

    write(join(sdir, 'S2_COMPLETE'))
    
    finish_str = f'''
=====================================
{get_time_date()}
Finished step 2: bedpostx
Arguments: 
{pprint.pformat(params, width=1)}
Total time: {get_time_string(time.time() - start_time)} (HH:MM:SS)
=====================================
'''
    write(stdout, finish_str)
    print(finish_str)



#!/usr/bin/env python3
import argparse,parsl,os,sys,glob,shutil
from os.path import *
from mappertrac.subscripts import *

script_dir = abspath(os.path.dirname(os.path.realpath(__file__)))
cwd = abspath(os.getcwd())

def parse_args(args):

    parser = argparse.ArgumentParser()

    if not '--test' in sys.argv:
        parser.add_argument('inputs', nargs='+',
            help='Paths to BIDS subject folder(s).')

    parser.add_argument('--test', action='store_true',
        help='Test using the example subject.')

    workflow_group = parser.add_mutually_exclusive_group(required=True)

    workflow_group.add_argument('--freesurfer', '--s1_freesurfer', '-s1', action='store_true',
        help='Run step 1: freesurfer.')

    workflow_group.add_argument('--bedpostx', '--s2_bedpostx', '-s2', action='store_true',
        help='Run step 2: bedpostx.')

    workflow_group.add_argument('--probtrackx', '--s3_probtrackx', '-s3', action='store_true',
        help='Run step 3: probtrackx.')

    parser.add_argument('--outputs', '-o', default='mappertrac_outputs/',
        help='Path to output directories.')

    parser.add_argument('--container', default=join(cwd, 'image.sif'),
        help='Path to Singularity container image.')

    parser.add_argument('--pbtx_sample_count', default=200,
        help='Number of probtrackx samples per voxel.')

    scheduler_group = parser.add_mutually_exclusive_group()

    scheduler_group.add_argument('--slurm', action='store_true',
        help='Use the Slurm scheduler.')

    scheduler_group.add_argument('--cobalt', action='store_true',
        help='Use the Cobalt scheduler.')

    scheduler_group.add_argument('--grid_engine', action='store_true',
        help='Use the Grid Engine scheduler.')

    parser.add_argument('--nnodes', '-n', default=1,
        help='Scheduler: number of nodes.')

    parser.add_argument('--bank', '-b', default='asccasc',
        help='Scheduler: bank to charge for jobs.')

    parser.add_argument('--partition', '-p', default='pbatch',
        help='Scheduler: partition to assign jobs.')

    parser.add_argument('--walltime', '-t', default='11:59:00',
        help='Scheduler: walltime in format HH:MM:SS.')

    return parser.parse_args()

def main():

    args = parse_args(sys.argv[1:])

    if shutil.which('singularity') is None:
        raise Exception(f"Missing Singularity executable in PATH.\n\n" +
            f"Please ensure Singularity is installed: https://sylabs.io/guides/3.0/user-guide/installation.html")

    if not exists(args.container):
        raise Exception(f"Missing container image at {abspath(args.container)}\n\n" +
            f"Either specify another image with --container\n\n" +
            f"Or build the container with the recipe at: {join(script_dir, 'data/container/recipe.def')}\n\n" +
            f"Or download the container at: https://www.dropbox.com/s/2v74hra04bo22w7/image.sif?dl=1\n")

    if args.test:
        args.inputs = join(script_dir, 'data/example_inputs/sub-*/')

    if isinstance(args.inputs, str):
        args.inputs = glob(args.inputs)

    output_dir = abspath(normpath(args.outputs))
    smart_mkdir(output_dir)

    all_params = []

    # Copy reads to subject directory
    session_dirs = []
    subject_dirs = [] # subjects without sessions
    for input_dir in args.inputs:
        input_dir = abspath(input_dir)

        sessions = glob(join(input_dir, 'ses-*/'))
        if len(sessions) > 0:
            session_dirs += [normpath(_) for _ in sessions]
        else:
            subject_dirs += [normpath(input_dir)]

    base_params = {
        'container': abspath(args.container),
        'script_dir': abspath(script_dir),
        'output_dir': output_dir,
        'pbtx_sample_count': int(args.pbtx_sample_count),
    }

    for input_dir in session_dirs:
        subject = basename(dirname(input_dir))
        session = basename(input_dir)
        subject_dir = join(output_dir, 'derivatives', subject, session)
        param = base_params.copy()
        param.update({
            'input_dir': input_dir,
            'work_dir': join(subject_dir, 'work_dir'),
            'ID': f'{subject}_{session}',
            'stdout': join(subject_dir, 'worker.stdout'),
        })
        all_params.append(param)

    for input_dir in subject_dirs:
        subject = basename(input_dir)
        subject_dir = join(output_dir, 'derivatives', subject)
        param = base_params.copy()
        param.update({
            'input_dir': input_dir,
            'work_dir': join(subject_dir, 'work_dir'),
            'ID': subject,
            'stdout': join(subject_dir, 'worker.stdout'),
        })
        all_params.append(param)

    if args.slurm:
        executor = parsl.executors.HighThroughputExecutor(
            label="worker",
            address=parsl.addresses.address_by_hostname(),
            provider=parsl.providers.SlurmProvider(
                args.partition,
                launcher=parsl.launchers.SrunLauncher(),
                nodes_per_block=int(args.nnodes),
                init_blocks=1,
                max_blocks=1,
                scheduler_options="#SBATCH --exclusive\n#SBATCH -A {}\n".format(args.bank),
                worker_init=f"export PYTHONPATH=$PYTHONPATH:{os.getcwd()}",
                walltime=args.walltime,
                move_files=False,
            ),
        )
    elif args.cobalt:
        executor = parsl.executors.HighThroughputExecutor(
            label="worker",
            worker_debug=True,
            address=parsl.addresses.address_by_hostname(),
            provider=parsl.providers.CobaltProvider(
                channel=parsl.channels.LocalChannel(),
                launcher=parsl.launchers.SimpleLauncher(),
                nodes_per_block=int(args.nnodes),
                init_blocks=1,
                max_blocks=1,
                scheduler_options=f"#SBATCH --exclusive\n#SBATCH -A {args.bank}\n",
                worker_init=f"export PYTHONPATH=$PYTHONPATH:{os.getcwd()}",
                # worker_init='source /home/madduri/setup_cooley_env.sh',
                walltime=args.walltime,
                account=args.bank,
                queue=args.partition,
            ),
        )
    elif args.grid_engine:
        executor = parsl.executors.HighThroughputExecutor(
            label="worker",
            worker_debug=True,
            address=parsl.addresses.address_by_hostname(),
            max_workers=int(cores_per_node[step]), # cap workers, or else defaults to infinity.
            provider=parsl.providers.GridEngineProvider(
                channel=parsl.channels.LocalChannel(),
                launcher=parsl.launchers.SingleNodeLauncher(),
                nodes_per_block=int(args.nnodes),
                init_blocks=1,
                max_blocks=1,
                scheduler_options="#SBATCH --exclusive\n#SBATCH -A {}\n".format(args.bank),
                worker_init=f"export PYTHONPATH=$PYTHONPATH:{os.getcwd()}",
                walltime=args.walltime,
                queue='gpu.q', # enables Wynton GPU queues
            ),
        )
    else:
        executor = parsl.executors.ThreadPoolExecutor(label="worker")

    config = parsl.config.Config(executors=[executor])
    parsl.set_stream_logger()
    parsl.load(config)

    if args.freesurfer:
        results =  []
        for params in all_params:
            results.append(run_freesurfer(params))
        for r in results:
            r.result()

    elif args.bedpostx:
        results =  []
        for params in all_params:
            results.append(run_bedpostx(params))
        for r in results:
            r.result()

    elif args.probtrackx:
        results =  []
        for params in all_params:
            results.append(run_probtrackx(params))
        for r in results:
            r.result()

if __name__ == '__main__':
    main()
